import math
import operator
from functools import reduce
from pathlib import Path

import cv2
import numpy as np
import pyclipper
from pyclipper import PyclipperOffset
from pydantic.dataclasses import dataclass
from pydantic.fields import List
from shapely.geometry import Polygon


@dataclass
class ImageTextCutterPointsModel:
    points: List[List]  # 图片中文字所在的位置,示例 [[1,1],[1,1],[1,1],[1,1]]
    label: str  # 文字内容


@dataclass
class ImageTextCutterModel:
    image_file_path: str  # 图片文件路径
    points: List[ImageTextCutterPointsModel]


@dataclass
class RotatedBoxModel:
    degree: int
    center_x: float
    center_y: float
    box_height: float
    box_width: float


class ImageTextCutter:
    def cut_image(self, cutter_config: List[ImageTextCutterModel]):
        """
        截取图片中的文字区域
        :param cutter_config: 切图配置
        :return:
        """
        for model in cutter_config:
            image_file_path = Path(model.image_file_path)
            image = cv2.imread(str(image_file_path), cv2.COLOR_BGR2RGB)
            for point_config in model.points:
                points_array = np.array(point_config.points).astype(np.int32)
                rotated_box = self.get_min_area_bbox(image, points_array)

                if rotated_box is None:
                    continue

                cropped_image = self.get_rotated_box_roi_from_image(image, rotated_box)

                yield cropped_image, point_config.label

    def rotate_points(self, points: List[List], degree=0, center=(0, 0)):
        """
        逆时针绕着一个点旋转点,points是非归一化的值
        :param points: 需要旋转的点
        :param degree: 角度
        :param center: 中心点
        :return:旋转后的点
        """
        angle = np.deg2rad(degree)  # 返回角度对应的数值，比对180°是3.14
        rotate_matrix = np.array([[np.cos(angle), -np.sin(angle)],
                                  [np.sin(angle), np.cos(angle)]])
        center = np.atleast_2d(center)  # 输入视为至少具有二维的数组。3.0 -> [[3.]]
        points = np.atleast_2d(points)
        return np.reshape((rotate_matrix @ (points.T - center.T) + center.T).T, (-1, 2))

    def get_coordinates_of_rotated_box(self, image: np.ndarray,
                                       rotated_box: RotatedBoxModel,
                                       scale_ratio=(1.0, 1.0)):
        """
        获取旋转的矩形的对应的四个顶点坐标
        :param image:对应的图像
        :param rotated_box: 旋转的矩形
        :param scale_ratio: 获取bbox的尺度，如果为tuple，则分别表示width和height的scale ratio，如果是一个数字就同时表示两个
        :return:    四个对应在图像中的坐标点
        """
        h, w = image.shape[:2]
        center_x = rotated_box.center_x
        center_y = rotated_box.center_y
        if isinstance(scale_ratio, tuple):
            width_ratio, height_ratio = scale_ratio
        else:
            width_ratio, height_ratio = scale_ratio, scale_ratio
        half_box_width = rotated_box.box_width * width_ratio / 2
        half_box_height = rotated_box.box_height * height_ratio / 2
        raw_points = np.array([
            [center_x - half_box_width, center_y - half_box_height],
            [center_x + half_box_width, center_y - half_box_height],
            [center_x + half_box_width, center_y + half_box_height],
            [center_x - half_box_width, center_y + half_box_height]
        ]) * (w, h)
        rotated_points = self.rotate_points(raw_points, rotated_box.degree, (center_x * w, center_y * h))
        rotated_points[:, 0] = np.clip(rotated_points[:, 0], a_min=0, a_max=w)
        rotated_points[:, 1] = np.clip(rotated_points[:, 1], a_min=0, a_max=h)
        return rotated_points.astype(np.int32)

    def get_rotated_box_roi_from_image(self, image: np.ndarray,
                                       rotated_box: RotatedBoxModel,
                                       scale_ratio=(1.0, 1.0)):
        """
        在图像中抠取一个旋转的box的roi
        :param image: 待抠取图像
        :param rotated_box: 旋转的box
        :param scale_ratio: 缩放比例，如果为tuple则分别为width和height的scale ratio，如果是数值的类型的，则width height的scale ratio一致
        """
        h, w = image.shape[:2]
        rotated_points = self.get_coordinates_of_rotated_box(image, rotated_box, scale_ratio)
        target_h = int(rotated_box.box_height * h)
        target_w = int(rotated_box.box_width * w)
        target_points = np.array([
            [0, 0],
            [target_w, 0],
            [target_w, target_h],
        ], dtype=np.float32)
        warp_matrix = cv2.getAffineTransform(rotated_points.astype(np.float32)[:3], target_points)
        cropped_image = cv2.warpAffine(image, warp_matrix, (target_w, target_h))
        return cropped_image

    def get_min_area_bbox(self, image, contour, scale_ratio=1.0) -> RotatedBoxModel:
        """
        获取一个contour对应的最小面积矩形
        note:主要是解决了旋转角度不合适的问题
        :param image: bbox所在图像
        :param contour:轮廓
        :param scale_ratio: 缩放比例
        :return 最小面积矩形的相关信息
        """
        h, w = image.shape[:2]
        if scale_ratio != 1:
            reshaped_contour = contour.reshape(-1, 2)
            current_polygon = Polygon(reshaped_contour)
            distance = current_polygon.area * scale_ratio / current_polygon.length
            offset = PyclipperOffset()
            offset.AddPath(reshaped_contour, pyclipper.JT_ROUND, pyclipper.ET_CLOSEDPOLYGON)
            all_paths = offset.Execute(distance)
            if len(all_paths) > 0:
                max_path = max(all_paths, key=lambda x: cv2.contourArea(np.array(x)))
                scaled_contour = np.array(max_path).reshape(-1, 1, 2)
            else:
                return None
        else:
            scaled_contour = contour
        try:
            # 会存在contour不合法的情况下，无法计算得到最小面积矩形
            rotated_box = cv2.minAreaRect(scaled_contour)
            if 45 <= rotated_box[2] <= 90:
                to_rotate_degree = rotated_box[2] - 90
                bbox_height, bbox_width = rotated_box[1]
            else:
                to_rotate_degree = rotated_box[2]
                bbox_width, bbox_height = rotated_box[1]
            # 几何信息归一化可以方便进行在缩放前的图像上进行操作
            rotated_box_model = RotatedBoxModel(
                degree=int(to_rotate_degree),
                center_x=rotated_box[0][0] / w,
                center_y=rotated_box[0][1] / h,
                box_height=bbox_height / h,
                box_width=bbox_width / w
            )
            return rotated_box_model
        except Exception:
            return None

    def clockwise_sort_points(self, point_coordinates) -> List[List]:
        """
        以左上角为起点的顺时针排序
        原理就是将笛卡尔坐标转换为极坐标，然后对极坐标的φ进行排序
        :param point_coordinates: 待排序的点[[x1,y1],[x2,y2],[x3,y3],[x4,y4]]
        :return: 排序完成的点
        """
        center_point = tuple(
            map(operator.truediv, reduce(lambda x, y: map(operator.add, x, y), point_coordinates),
                [len(point_coordinates)] * 2))
        return sorted(point_coordinates, key=lambda coord: (180 + math.degrees(
            math.atan2(*tuple(map(operator.sub, coord, center_point))[::-1]))) % 360)
